////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2018 The University of North Carolina at Chapel Hill.
// All Rights Reserved.
//
// Permission to use, copy, modify and distribute this software and its
// documentation for educational, research and non-profit purposes, without
// fee, and without a written agreement is hereby granted, provided that the
// above copyright notice and the following three paragraphs appear in all
// copies.
//
// IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
// LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
// USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
// OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGES.
//
// THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
// PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
// NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// The authors may be contacted via:
//
// US Mail: Real-Time Systems Group at UNC
// Department of Computer Science
// Sitterson Hall
// University of N. Carolina
// Chapel Hill, NC 27599-3175
//
// EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu
//
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "cycles.h"
#include "crnlp_general_fgls.h"
#include "nesting.h"
#include "mem.h"
#include "queue.h"

void crnlp_general_fgls_init(crnlp_general_fgls_lock_t* lock)
{
    int i = 0;

    // Initialize resource queues
    for (i = 0; i < RES_COUNT; i++)
    {
        lock->queues[i].head = NULL;
        lock->queues[i].tail = NULL;
    }

    // Initialize the requests
    for (i = 0; i < QUEUE_LENGTH; i++)
    {
        lock->requests[i].requested = 0;
        lock->requests[i].length = 0;
        lock->requests[i].request_id = i;
        lock->requests[i].enabled = 0;
        lock->requests[i].needs_service = 0;
    }

    lock->pending_requests = 0;

    lock->service_index = 0;
    lock->server_exists = 0;
}

uint64_t *crnlp_general_fgls_lock(crnlp_general_fgls_lock_t* lock, uint8_t ind)
{
    crnlp_general_fgls_request_t *self, *other;
    volatile crnlp_general_fgls_queue_t *queue, *queue2;
    volatile struct crnlp_general_fgls_node_t *node, *next_node;

    int r, s;
    uint8_t start_in_sched, i, j;
    uint8_t is_lock_server;
    unsigned long one, tmp, tmp2;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    one = 1;

    self = &(lock->requests[ind]);

    self->enabled = 0;

    // Mark this request as needing to be serviced for a lock
    self->done = 0;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then we switch to measuring
    // blocking time (the request must have been serviced)
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[0] += end - start;
        start = get_cycles();
    }

    // Wait until this request is satisfied, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will eventually become
    // satisfied, after potentially servicing other requests)
    while (!is_lock_server && self->enabled == 0) 
    {
        // Try to become the lock server (in order to help service requests
        // while waiting to be satisfied)
        if (!lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
        }
    }

    // If this core is not the lock server, then the request must be satisfied
    if (!is_lock_server)
    {
        end = get_cycles();
        ret[1] += end - start;

        return ret;
    }

    // Otherwise, help out as long as this request has not yet been serviced
    // and is not yet satisfied
    while (self->needs_service != 0 || self->enabled == 0)
    {
        // Check if the next request needs to be serviced
        i = lock->service_index;
        if (lock->requests[i].needs_service == 1)
        {
            // Service request i
            other = &(lock->requests[i]);
            if (other->done == 0)
            {
                // Lock code
                uint64_t start_in_sched, prev_start, potential_start;

                // Find the earliest possible start time
                start_in_sched = get_cycles();
                prev_start = start_in_sched;

                while (1)
                {
                    // Loop through the required resources
                    tmp = other->requested;
                    r = ffsl(tmp);
                    while (r != 0)
                    {
                        r = r - 1;

                        // For this resource, find the earliest possible start time after
                        // the current start_in_sched within the resource queue
                        queue = &(lock->queues[r]);

                        if (queue->head)
                        {
                            // If this request cannot be satisfied before the first node,
                            // we need to find the node it will follow
                            if ((queue->head->request->enabled == 1) ||
                                (get_cycles() + other->length >= queue->head->start_time))
                            {
                                node = queue->head->next;

                                while (node)
                                {
                                    potential_start = node->prev->start_time +
                                                      node->prev->request->length;
                                    potential_start = potential_start < start_in_sched ?
                                                          start_in_sched :
                                                          potential_start;
                                    if (potential_start + other->length < node->start_time)
                                    {
                                        // This location is delay-preserving!
                                        start_in_sched = potential_start;
                                        break;
                                    }
                                    else
                                    {
                                        // We need to continue looking in the schedule
                                        node = node->next;
                                    }
                                }

                                // If node is NULL, we belong at the tail of the queue
                                if (!node)
                                {
                                    potential_start = queue->tail->start_time +
                                                      queue->tail->request->length;
                                    if (potential_start > start_in_sched)
                                    {
                                        start_in_sched = potential_start;
                                    }
                                }
                            }
                        }

                        tmp &= ~(one << r);
                        r = ffsl(tmp);
                    }

                    if (prev_start == start_in_sched)
                    {
                        // If the selected time did not change, we are done looping
                        break;
                    }
                    else
                    {
                        // Otherwise, update prev_start and go around again
                        prev_start = start_in_sched;
                    }
                }

                // Update pointers to insert the request into the queue for each
                // required resource
                j = 0;
                int is_head = 1;
                tmp = other->requested;
                r = ffsl(tmp);
                while (r != 0)
                {
                    r = r - 1;

                    // Grab a queue from the pool
                    node = &(lock->nodes[i][j++]);

                    node->request = other;
                    node->start_time = start_in_sched;

                    node->prev = NULL;
                    node->next = NULL;

                    // Insert this node into the queue
                    queue = &(lock->queues[r]);
                    if (!queue->head)
                    {
                        queue->head = node;
                        queue->tail = node;
                    }
                    else
                    {
                        // Find the node it will precede
                        next_node = queue->head;
                        while (next_node && (next_node->start_time <= start_in_sched))
                        {
                            next_node = next_node->next;
                        }

                        // Insert the new node before next_node
                        if (!next_node)
                        {
                            queue->tail->next = node;
                            node->prev = queue->tail;

                            queue->tail = node;

                            is_head = 0;
                        }
                        else if (next_node == queue->head)
                        {
                            queue->head = node;

                            next_node->prev = node;
                            node->next = next_node;
                        }
                        else
                        {
                            next_node->prev->next = node;
                            node->prev = next_node->prev;

                            next_node->prev = node;
                            node->next = next_node;

                            is_head = 0;
                        }
                    }

                    tmp &= ~(one << r);
                    r = ffsl(tmp);
                }

                if (is_head == 1)
                {
                    other->enabled = 1;
                }

                // Update the count of pending requests
                lock->pending_requests += 1;

                mb();

                // Mark this request as serviced
                other->needs_service = 0;
            }
            else
            {
                // Unlock code

                // Iterate through the resource queues for this request
                tmp = other->requested;
                r = ffsl(tmp);
                while (r != 0)
                {
                    r = r - 1;

                    queue = &(lock->queues[r]);
                    node = queue->head;

                    // Update the head
                    queue->head = queue->head->next;

                    // Update the new head's prev pointer
                    if (queue->head)
                    {
                        queue->head->prev = NULL;
                    }
                    else
                    {
                        queue->tail = NULL;
                    }

                    tmp &= ~(one << r);
                    r = ffsl(tmp);
                }
                // Check the heads of the resource queues to determine
                // if they are now enabled
                tmp = other->requested;
                r = ffsl(tmp);
                while (r != 0)
                {
                    r = r - 1;

                    queue = &(lock->queues[r]);

                    // If the queue is empty, skip it
                    if (queue->head)
                    {
                        // Check if the head of this queue is now enabled,
                        // by checking the queues for this request
                        int is_enabled = 1;

                        tmp2 = queue->head->request->requested;
                        s = ffsl(tmp2);
                        while (s != 0)
                        {
                            s = s - 1;

                            queue2 = &(lock->queues[s]);

                            // If this request is not the head, it can't
                            // be enabled
                            if (queue->head->request != queue2->head->request)
                            {
                                is_enabled = 0;
                                break;
                            }

                            tmp2 &= ~(one << s);
                            s = ffsl(tmp2);
                        }

                        // If the request at the head is enabled, set
                        // its bit
                        if (is_enabled == 1)
                        {
                            queue->head->request->enabled = 1;
                        }
                    }

                    tmp &= ~(one << r);
                    r = ffsl(tmp);
                }

                // Decrement the pending request count, as this
                // request is now finished
                lock->pending_requests -= 1;

                mb();

                // Mark this request as serviced
                other->needs_service = 0;
            }

            // If this request was just serviced, then we switch to measuring
            // blocking time
            if (i == ind)
            {
                end = get_cycles();
                ret[0] += end - start;
                start = get_cycles();
            }
        }

        // Move onto the next request
        lock->service_index = (i + 1) % QUEUE_LENGTH;
    }

    // Set the server to not exist
    lock->server_exists = 0;

    end = get_cycles();
    ret[1] += end - start;
    start = get_cycles();

    // Satisfied!
    return ret;
}

uint64_t *crnlp_general_fgls_unlock(crnlp_general_fgls_lock_t* lock, uint8_t ind)
{
    crnlp_general_fgls_request_t *self, *other;
    volatile crnlp_general_fgls_queue_t *queue, *queue2;
    volatile struct crnlp_general_fgls_node_t *node, *next_node;
    uint8_t is_lock_server;

    int r, s;
    unsigned long one, tmp, tmp2;

    uint64_t *ret;
    uint64_t start, end;

    ret = (uint64_t *)malloc(sizeof(uint64_t) * 2);
    ret[0] = ret[1] = 0;

    start = get_cycles();

    one = 1;

    self = &(lock->requests[ind]);

    // Mark this request as needing to be serviced for an unlock
    self->done = 1;
    mb();
    self->needs_service = 1;

    // Check whether a lock server exists (and potentially become it)
    is_lock_server = 0;
    if (!lock->server_exists)
    {
        is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
    }

    // Wait until this request is serviced, or this core becomes the lock server
    // (if it is already the lock server, skip this; it will service itself later)
    while (!is_lock_server && self->needs_service != 0)
    {
        // Try to become the lock server (in the case the previous lock server
        // completed without servicing this request)
        if (!lock->server_exists)
        {
            is_lock_server = __sync_bool_compare_and_swap(&(lock->server_exists), 0, 1);
        }
    }

    // If this core is the lock server and has not been serviced, be
    // greedy and just service this unlock
    if (is_lock_server && self->needs_service == 0)
    {
        // Oops, this core shouldn't be the server
        lock->server_exists = 0;
    }
    else if (is_lock_server)
    {
        // Iterate through the resource queues for this request
        tmp = self->requested;
        r = ffsl(tmp);
        while (r != 0)
        {
            r = r - 1;

            queue = &(lock->queues[r]);
            node = queue->head;

            // Update the head
            queue->head = queue->head->next;

            // Update the new head's prev pointer
            if (queue->head)
            {
                queue->head->prev = NULL;
            }
            else
            {
                queue->tail = NULL;
            }

            tmp &= ~(one << r);
            r = ffsl(tmp);
        }

        // Check the heads of the resource queues to determine
        // if they are now enabled
        tmp = self->requested;
        r = ffsl(tmp);
        while (r != 0)
        {
            r = r - 1;

            queue = &(lock->queues[r]);

            // If the queue is empty, skip it
            if (queue->head)
            {
                // Check if the head of this queue is now enabled,
                // by checking the queues for this request
                int is_enabled = 1;

                tmp2 = queue->head->request->requested;
                s = ffsl(tmp2);
                while (s != 0)
                {
                    s = s - 1;

                    queue2 = &(lock->queues[s]);

                    // If this request is not the head, it can't
                    // be enabled
                    if (queue->head->request != queue2->head->request)
                    {
                        is_enabled = 0;
                        break;
                    }

                    tmp2 &= ~(one << s);
                    s = ffsl(tmp2);
                }

                // If the request at the head is enabled, set
                // its bit
                if (is_enabled == 1)
                {
                    queue->head->request->enabled = 1;
                }
            }

            tmp &= ~(one << r);
            r = ffsl(tmp);
        }

        // Decrement the pending request count, as this
        // request is now finished
        lock->pending_requests -= 1;

        mb();

        // Mark this request as serviced
        self->needs_service = 0;

        // Set the server to not exist
        lock->server_exists = 0;
    }

    end = get_cycles();
    ret[0] += end - start;

    return ret;
}

